### Calculating
library(tidyverse)


### Auxiliar functions

ymtd_logistic <- function(rho, eta, cov, theta, x){
  
  out <- ((qlogis(theta) - qlogis(rho[1])) - 
            (qlogis(rho[2]) - qlogis(rho[1]))*x - 
            (qlogis(rho[4]) - qlogis(rho[2]))*cov)/
    ((qlogis(rho[3]) - qlogis(rho[1])) + eta*x)
  
  return(out)
}

mtd_doses <- function(cov, rho, eta, theta, nx, ny, grid) {
  
  # special senarios
  # do not recommend MTD if rho00 > theta + 0.1
  if(rho[1] > theta + 0.1) {
    doses = data.frame(x = NA, y = NA, cov = NA)
  } else if(ymtd_logistic(rho, eta, cov, theta, 0) < 0) {
    doses = data.frame(x = 0, y = 0, cov = cov)
  } else if(ymtd_logistic(rho, eta, cov, theta, 1) > 1) {
    doses = data.frame(x = 1, y = 1, cov)
  } else {
    
    # set number of descrete levels for drug X and Y
    xgrid = unique(grid$x)
    ygrid = unique(grid$y)
    
    # find intercept
    xmin = 0
    xmax = 1
    
    aux_xmax <- function(x, rho, eta, cov, theta) {
      out <- ymtd_logistic(rho, eta, cov, theta, x)
    }
    
    aux_xmin <- function(x, rho, eta, cov, theta) {
      out <- ymtd_logistic(rho, eta, cov, theta, x) - 1
    }
    
    if(ymtd_logistic(rho, eta, cov, theta, 0) > 1) 
      xmin = uniroot(aux_xmin, rho = rho, eta = eta, cov = cov,  
                     theta = theta, 
                     interval = c(0, 1), extendInt="downX")$root
    if(ymtd_logistic(rho, eta, cov, theta, 1) < 0) 
      xmax = uniroot(aux_xmax, rho = rho, eta = eta, cov = cov,  
                     theta = theta, 
                     interval = c(0,1), extendInt="downX")$root
    
    # Define the function to be minimized
    fdist <- function(x, pt, rho, eta, cov, theta){
      # pt is the point
      d <- (pt[1] - x)^2 + 
        (pt[2]- ymtd_logistic(rho, eta, cov, theta, x))^2
      
      out <- d^0.5
      return(out)
    }
    
    ######
    # calculate the mimimum distance for each dose combinations
    ld = matrix(nrow = length(xgrid), ncol = length(ygrid))
    
    for (i in 1:length(xgrid)) { 
      for (j in 1:length(ygrid)) {
        aal <- optimize(f=fdist, c(xmin,xmax), tol=0.0001, c(xgrid[i],ygrid[j]), 
                        rho, eta, cov, theta)
        ld[i,j] <- aal$objective
      }
    }
    
    ######
    # find the best combinations
    
    # 1. select x fist
    ldx <- ld
    for (i in 1:length(xgrid)) 
      ldx[i,which(ldx[i,] != min(ldx[i,], na.rm=T))] <- NA
    
    for (j in 1:length(ygrid)) {
      if (sum(!is.na(ldx[,j])) > 0)
        ldx[which(ldx[,j] != min(ldx[,j], na.rm=T)), j] <- NA
    }
    
    # 2. select y fist
    ldy <- ld
    for (j in 1:length(ygrid)) 
      ldy[which(ldy[,j] != min(ldy[,j], na.rm=T)), j] <- NA
    
    for (i in 1:length(xgrid)) {
      if (sum(!is.na(ldy[i,])) > 0)
        ldy[i,which(ldy[i,] != min(ldy[i,], na.rm=T))] <- NA
    }
    
    # intersection ld1 and ld2
    temp <- data.frame(ldx = as.numeric(ldx), ldy = as.numeric(ldy))
    ld <- matrix(apply(temp[, 1:2], 1, mean), nrow = nx, ncol = ny) # intersection
    ld[is.nan(ld)] <-  NA
    
    # determine the dose combinations for MTD
    dosex <- matrix(xgrid, nrow = length(xgrid), ncol = length(ygrid))
    dosey <- matrix(ygrid, byrow = T, nrow = length(xgrid), ncol = length(ygrid))
    doses <- data.frame(x = dosex[which(!is.na(ld))], 
                        y = dosey[which(!is.na(ld))],
                        cov = cov)
    
    #doses is the dose combinations that are close to the true MTD curve
    return(doses)
  }
}

mtd_selection <- function(rho_sim, eta_sim, theta, 
                          post_dlt_sim, post_low, grid, epsilon, 
                          stop = 0.8, mtd_limit = 0.1,
                          file = "mtd_selection") {
  
  mtd_subset <- function(mtd){
    out <- ifelse(all(mtd == 1), 1, 0)
    return(out)
  }
  
  mtd_length <- function(mtd){
    out <- sum(mtd)
    return(out)
  }
  
  mtd_proportion <- function(mtd){
    out <- sum(mtd)/length(mtd)
    return(out)
  }
  
  nx <- nlevels(as.factor(grid$x))
  ny <- nlevels(as.factor(grid$y))
  xgrid = unique(grid$x)
  ygrid = unique(grid$y)
  M <- nrow(rho_sim)
  
  # recommend dose combination based on estimated MTD curve
  
  result = data.frame(x=NA, y=NA, cov = NA, trial=NA)
  for (i in 1:nrow(xdose_sim)) {
    
    recommended_mtd0 <- mtd_doses(cov = 0, 
                                  rho = rho_sim[i, ], eta = eta_sim[i], 
                                  theta = theta, nx = nx, ny = ny, 
                                  grid = grid)
    recommended_mtd1 <- mtd_doses(cov = 1, 
                                 rho = rho_sim[i, ], eta = eta_sim[i], 
                                 theta = theta, nx = nx, ny = ny, 
                                 grid = grid)
    recommended_mtd <- rbind(recommended_mtd0, recommended_mtd1)
    
    #################
    # postdlt control, added 03/06/2015
    temp <- post_dlt_sim %>% filter(trial == i)
    aux <- left_join(recommended_mtd, temp, by = c("x", "y", "cov"))
    
    # If postdlt > delta5, exclude those dose combinations
    aux <- aux %>% filter(p < epsilon) %>% select(x, y, cov, trial)
    ################
    
    # do not recommend MTD if any postlow >= 0.8
    if (max(stop_sim[i, ]) < stop) 
      result <- rbind(result, aux)
  }
  
  # remove NAs (including NA in the first row)
  result = result[!is.na(result$x),]
  
  ######################################################################
  # Percent of MTD individual selection
  ######################################################################
  
  individual <- result %>% group_by(cov, x, y) %>% 
    count() %>% 
    ungroup()  %>%
    complete(cov, x, y, fill = list(n = 0)) %>% group_by(cov) %>%
    mutate(percent = round(100*n/sum(n), 1)) %>% ungroup()
    
  
  grid$mtd <- as.factor(ifelse(abs(grid$p - round(theta, 2)) <= mtd_limit, 1, 0))
  gp <- ggplot(grid, aes(x = x, y = y, label = individual$percent, colour = mtd)) + 
    geom_text() + theme_bw() + facet_grid(. ~ cov) +
    labs(x = "Dose (Drug A)", y = "Dose (Drug B)") +
    scale_x_continuous(breaks = round(unique(grid$x), 2)) +
    scale_y_continuous(breaks = round(unique(grid$y), 2)) +
    scale_colour_manual(values = c("black", "blue")) +
    theme(panel.grid.major.x = element_line(size = 1/nx), 
          panel.grid.major.y = element_line(size = 1/ny),
          legend.position = "none") + ggsave(file = paste0(file, ".tiff"))
  
  
  ######################################################################
  # Percent of MTD subset selection
  ######################################################################
  aux <- left_join(result, grid)
  aux$trial <- as.factor(aux$trial)
  aux$cov <- as.factor(aux$cov)
  aux$mtd <- as.numeric(aux$mtd) - 1
  
  ps <- rep(NA, 2)
  temp <- tapply(aux$mtd[aux$cov == 0], aux$trial[aux$cov == 0], mtd_subset)
  ps[1] <- 100*sum(temp, na.rm = TRUE)/nrow(rho_sim)
  temp <- tapply(aux$mtd[aux$cov == 1], aux$trial[aux$cov == 1], mtd_subset)
  ps[2] <- 100*sum(temp, na.rm = TRUE)/nrow(rho_sim)
  
  sk <- matrix(NA, nrow = 2, ncol = 3)
  skk0 <- tapply(aux$mtd[aux$cov == 0], aux$trial[aux$cov == 0], mtd_length)
  temp <- cumsum(rev(100*table(skk0)/nrow(rho_sim)))
  temp <- temp[names(temp) %in% c("3", "2", "1")]
  if(length(temp) < 3) temp <- c(rep(NA, 3 - length(temp)), temp)
  sk[1, ] <- temp

  skk1 <- tapply(aux$mtd[aux$cov == 1], aux$trial[aux$cov == 1], mtd_length)
  temp <- cumsum(rev(100*table(skk1)/nrow(rho_sim)))
  temp <- temp[names(temp) %in% c("3", "2", "1")]
  if(length(temp) < 3) temp <- c(rep(NA, 3 - length(temp)), temp)
  sk[2, ] <- temp
  
  avw <- rep(NA, 2)
  avw[1] <- 100*sum(skk0, na.rm = TRUE)/length(aux$mtd[aux$cov == 0])
  avw[2] <- 100*sum(skk1, na.rm = TRUE)/length(aux$mtd[aux$cov == 1])
  
  out <- list(subset = ps, length = sk, 
              w.proportion = avw, plot = gp)
  return(out)
}

mtd_selection(rho_sim, eta_sim, theta, 
                          post_dlt_sim, post_low, grid, epsilon = 0.7, 
                          stop = 0.8, mtd_limit = 0.1,
                          file = "mtd_selection")

mean(rowMeans(dlt_sim))
mean(rowMeans(dlt_sim) > theta + 0.1)
mean(rowMeans(dlt_sim) < theta - 0.1)

dlt_sim <- dlt
cov_sim <- Zcov

mean(unlist(map2(.x = split(dlt_sim, 1:nrow(dlt_sim)), 
                 .y = split(cov_sim, 1:nrow(cov_sim)), 
                 .f = ~ mean(.x[.y == 0]))))
mean(unlist(map2(.x = split(dlt_sim, 1:nrow(dlt_sim)), 
                 .y = split(cov_sim, 1:nrow(cov_sim)), 
                 .f = ~ mean(.x[.y == 0]))) > theta + 0.1)
mean(unlist(map2(.x = split(dlt_sim, 1:nrow(dlt_sim)), 
                 .y = split(cov_sim, 1:nrow(cov_sim)), 
                 .f = ~ mean(.x[.y == 0]))) < theta - 0.1)

mean(unlist(map2(.x = split(dlt_sim, 1:nrow(dlt_sim)), 
                 .y = split(cov_sim, 1:nrow(cov_sim)), 
                 .f = ~ mean(.x[.y == 1]))))
mean(unlist(map2(.x = split(dlt_sim, 1:nrow(dlt_sim)), 
                 .y = split(cov_sim, 1:nrow(cov_sim)), 
                 .f = ~ mean(.x[.y == 1]))) > theta + 0.1)
mean(unlist(map2(.x = split(dlt_sim, 1:nrow(dlt_sim)), 
                 .y = split(cov_sim, 1:nrow(cov_sim)), 
                 .f = ~ mean(.x[.y == 1]))) < theta - 0.1)
