library(parallel)
library(doParallel)
library(R2OpenBUGS)
library(rjags)

############## Parameters
#Number of cores
ncores <- 10

# 01, 02, 03, 04
scenario <- 01

#logistic, cloglog, probit, normal
model <- "cloglog"

# Trial
sample_size <- 20
npatients <- 2
nc <- sample_size/npatients

n_sim <- 2000
n_adapt <- 5000
burn_in <- 10000
n_chains <- 1
n_thin <- 1
n_mcmc <- 5000

rho_prior <- matrix(1, ncol = 2, nrow = 4)
interaction_prior <- matrix(c(0.8, 0.0384), ncol = 2, nrow = 1)

theta <- 0.33
alpha <- 0.25
limits_xdose <- c(0, 1)
limits_ydose <- c(0, 1)
deltax <- 0.2
deltay <- 0.2

stop01 <- 0.1
stop02 <- 0.3
min.patients <- 1

# Loading scenario
source(paste0("scenario0", scenario, ".R"))

############## Auxiliar functions

comb <- function(...) {
  mapply('rbind', ..., SIMPLIFY=FALSE)
}

response_binary2d <- function(x, y, cov, rho, eta, model, theta){
  
  beta <- rep(NA, 5)
  beta[1] <- qlogis(rho[1])
  beta[2] <- qlogis(rho[2]) - qlogis(rho[1])
  beta[3] <- qlogis(rho[3]) - qlogis(rho[1])
  beta[4] <- qlogis(rho[4]) - qlogis(rho[2])
  beta[5] <- eta
  design <- cbind(1, x, y, cov, x*y)
  
  p <- plogis(design%*%beta)
  
  if (model == "probit"){
    beta[1] <- beta[1] - qlogis(theta) + qnorm(theta)
    p <- pnorm(design%*%beta)
  }
  if (model == "normal"){
    beta[1] <- beta[1] - qlogis(theta) + qnorm(theta, 0, 2)
    p <- pnorm(design%*%beta, 0, 2)
  }
  if (model == "cloglog"){
    beta[1] <- beta[1] - qlogis(theta) + log(-log(1-theta))
    p <- 1 - exp(-exp(design%*%beta))
  }
  
  out <- rbinom(length(x), 1, p)
  
  return(out)
}

pdlt_extended2d <- function(x, y, rho, eta, model, theta){
  
  aux_pdlt <- function(parm, x, y, model, theta){
    rho <- parm[1:3]
    eta <- parm[4]
    
    beta <- rep(NA, 4)
    beta[1] <- qlogis(rho[1])
    beta[2] <- qlogis(rho[2]) - qlogis(rho[1])
    beta[3] <- qlogis(rho[3]) - qlogis(rho[1])
    beta[4] <- eta
    design <- cbind(1, x, y, x*y)
    
    p <- plogis(design%*%beta)
    
    return(p)
  }
  
  parm <- cbind(rho, eta)
  out <- apply(parm, 1, aux_pdlt, 
               x = x, y = y, model = model, theta = theta)
  
  return(out)
}

### JAGS
extended2d_jags <- function(data) {
  
  theta <- data$theta
  lbx <- - data$limits_xdose[1]/(data$limits_xdose[2] - data$limits_xdose[1])
  lby <- - data$limits_ydose[1]/(data$limits_ydose[2] - data$limits_ydose[1])
  rho_prior <- data$rho_prior
  interaction_prior <- data$interaction_prior
  design_matrix <- data$design_matrix
  dlt <- data$response[, 2]
  npatients <- data$response[, 1]
  
  x.index <- data$x.index
  y.index<- data$y.index
  
  n_adapt <- data$n_adapt
  burn_in <- data$burn_in
  n_mcmc <- data$n_mcmc
  n_thin <- data$n_thin
  n_chains <- data$n_chains
  
  # JAGS model function
  jfun <- "model {
  
  for(i in 1:nobs) {
  dlt[i] ~ dbin(p[i], 1)
  p[i] <- ifelse(1/(1 + exp(-lp[i])) == 1, 0.99, 1/(1 + exp(-lp[i])))
  lp[i] <- inprod(design_matrix[i, ], beta)
  }
  
  beta[1] <- logit(rho[1])
  beta[2] <- logit(rho[2]) - logit(rho[1])
  beta[3] <- logit(rho[3]) - logit(rho[1])
  beta[4] <- eta
  
  rho[1] <- limit*v[1]
  v[1] ~ dbeta(rho_prior[1, 1], rho_prior[1, 2])
  limit <- min(rho[2], rho[3])
  
  rho[2] <- v[2]
  rho[3] <- v[3]
  
  v[2] ~ dbeta(rho_prior[2, 1], rho_prior[2, 2])
  v[3] ~ dbeta(rho_prior[3, 1], rho_prior[3, 2])
  eta ~ dgamma(interaction_prior[1, 1], interaction_prior[1, 2])
}"
  
  data_base <- list('dlt' = dlt, 'design_matrix' = design_matrix, 
                    'nobs' = length(dlt), 'rho_prior' = rho_prior,
                    'interaction_prior' = interaction_prior)
  
  inits <- function() {
    v <- rep(NA, 3)
    v[1] <- rbeta(1, rho_prior[1, 1], rho_prior[1, 2])
    v[2] <- rbeta(1, rho_prior[2, 1], rho_prior[2, 2])
    v[3] <- rbeta(1, rho_prior[3, 1], rho_prior[3, 2])
    eta <- rgamma(1, interaction_prior[1, 1], interaction_prior[1, 2])
    out <- list(v = v, eta = eta)
    return(out)
  }
  
  # Calling JAGS
  j <- jags.model(textConnection(jfun),
                  data = data_base,
                  inits = inits(),
                  n.chains = n_chains,
                  n.adapt = n_adapt)
  update(j, burn_in)
  result <- coda.samples(j, variable.names = c("rho", "eta", "limit"),
                         n.iter = n_mcmc, thin = n_thin,
                         n.chains = n_chains)[[1]]
  rho <- result[, 3:5]
  limit <- result[, 2]
  eta <- result[, 1]
  
  out <- list(rho = rho, eta = eta, limit = limit)
  
  return(out)
}

standard_dose <- function(dose, min_dose, max_dose) {
  
  out <- (dose - min_dose)/(max_dose - min_dose)
  return(out)
}

nd_ewoc_extended2d <- function(data, x, y, cov, x.index, y.index, deltax, deltay){
  
  mcmc_sample <- extended2d_jags(data) 
  rho <- mcmc_sample$rho
  eta <- mcmc_sample$eta
  lbx <- - data$limits_xdose[1]/(data$limits_xdose[2] - data$limits_xdose[1])
  lby <- - data$limits_ydose[1]/(data$limits_ydose[2] - data$limits_ydose[1])
  
  ymtd <- ((qlogis(theta) - qlogis(rho[, 1])) - 
    (qlogis(rho[, 2]) - qlogis(rho[, 1]))*x[y.index])/
    ((qlogis(rho[, 3]) - qlogis(rho[, 1])) + eta*x[y.index])
  ymtd <- ymtd[ymtd > lby]
  
  next_ydose <- quantile(ymtd, probs = data$alpha)
  if ((next_ydose - y[y.index]) > deltay)
    next_ydose <- y[y.index] + deltay
  
  if (next_ydose > 1)
    next_ydose <- 1
  if (next_ydose < 0)
    next_ydose <- 0
  
  xmtd <- ((qlogis(theta) - qlogis(rho[, 1])) - 
    (qlogis(rho[, 3]) - qlogis(rho[, 1]))*y[x.index])/
    ((qlogis(rho[, 2]) - qlogis(rho[, 1])) + eta*y[x.index])
  xmtd <- xmtd[xmtd > lbx]
  
  next_xdose <- quantile(xmtd, probs = data$alpha)
  if ((next_xdose - x[x.index]) > deltax)
    next_xdose <- x[x.index] + deltax
  
  if (next_xdose > 1)
    next_xdose <- 1
  if (next_xdose < 0)
    next_xdose <- 0
  
  pdlt <- pdlt_extended2d(x = 0, y = 0, rho = rho, eta = eta, 
                model = "logistic", theta = theta)
  
  out <- list(next_ydose = next_ydose, next_xdose = next_xdose,
              ymtd = ymtd, xmtd = xmtd, rho = rho, eta = eta, 
              pdlt = pdlt)
}

### Simulation - Cov = 0
set.seed(1234)

n_dose <- sample_size + 1
xdose_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
ydose_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
cov_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
stop_sim <- matrix(NA, ncol = (sample_size-1), nrow = n_sim)
dlt_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
eta_sim <- matrix(NA, ncol = 1, nrow = n_sim)
time_sim <- matrix(NA, ncol = 1, nrow = n_sim)
rho_sim <- matrix(NA, ncol = 3, nrow = n_sim)
alpha_sim <- matrix(NA, ncol = n_dose, nrow = n_sim)
alpha_sim[, 1] <- alpha

registerDoParallel(ncores)
result <-
  foreach(i = 1:n_sim,
          .combine='comb',
          .multicombine=TRUE,
          .errorhandling= "remove",
          .init=list(list(), list(), list(), list(), list(), list(), list()),
          .packages='rjags') %dopar% {
            
            ## Binary Covariate
            cov <- matrix(rep(0, sample_size), ncol = 2)
           
            x.index <- 1
            y.index <- 2
            
            dlt <- matrix(NA, nrow = nc, ncol = npatients)
            dlt[1, ] <- 0
            xdose <- matrix(NA, nrow = nc, ncol = npatients)
            xdose[1, ] <- 0
            ydose <- matrix(NA, nrow = nc, ncol = npatients)
            ydose[1, ] <- 0
            stop_rule <- matrix(NA, nrow = (nc - 1), ncol = 2)
            
            for (j in 2:nc) {
              response <- data.frame(npatients = 1,
                                     dlt = as.numeric(t(dlt[1:(j-1), ])))
              design_matrix <- data.frame(intercept = 1,
                                          x = as.numeric(t(xdose[1:(j-1), ])),
                                          y = as.numeric(t(ydose[1:(j-1), ])), 
                                          xy = as.numeric(t(xdose[1:(j-1), ]))*
                                            as.numeric(t(ydose[1:(j-1), ])))
              
              x.index <- setdiff(1:2, x.index)
              y.index <- setdiff(1:2, x.index)
              
              my_data <- list(response = response, design_matrix = design_matrix,
                              theta = theta, alpha = alpha_sim[i, (j-1)],
                              limits_xdose = limits_xdose, 
                              limits_ydose = limits_ydose, 
                              rho_prior = rho_prior, 

                              interaction_prior = interaction_prior,
                              n_adapt = n_adapt, burn_in = burn_in, 
                              n_mcmc = n_mcmc, n_thin = n_thin, 
                              n_chains = n_chains,
                              y.index = y.index, x.index = x.index)
              
              dose <- nd_ewoc_extended2d(my_data, 
                                       x = xdose[(j-1), ],
                                       y = ydose[(j-1), ],
                                       cov = cov[(j-1), ],
                                       x.index = x.index,
                                       y.index = y.index,
                                       deltax = deltax,
                                       deltay = deltay)
              
              count <- table(cov[1:(j-1), ])
              cond <- names(count) == cov[j, x.index]
              
              xdose[j, x.index] <- dose$next_xdose
              xdose[j, y.index] <- xdose[(j-1), y.index]

              ydose[j, y.index] <- dose$next_ydose
              ydose[j, x.index] <- ydose[(j-1), x.index]
              
              alpha_sim[i, j] <- ifelse(alpha_sim[i, (j-1)] < 0.49, 
                                        alpha_sim[i, (j-1)] + 0.05, 0.5)
              
              dlt[j, ] <- response_binary2d(x = xdose[j, ], y = ydose[j, ], 
                                            cov = cov[j, ],
                                            rho = true_rho, eta = true_eta,
                                            model = model, theta = theta)
              
              stop_rule[(j-1), ] <- mean(dose$pdlt > theta + stop01) 
            }
  
  list(xdose = as.numeric(t(xdose)), 
       ydose = as.numeric(t(ydose)), 
       dlt = as.numeric(t(dlt)), 
       rho_sim = apply(dose$rho, 2, median), 
       eta_sim = median(dose$eta),
       cov = as.numeric(t(cov)),
       stop_rule =  as.numeric(t(stop_rule)))
          }
stopImplicitCluster()

xdose_sim0 <- t(apply(result[[1]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
ydose_sim0 <- t(apply(result[[2]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
dlt_sim0 <- t(apply(result[[3]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
rho_sim0 <- t(apply(result[[4]], 1, function (u) matrix(unlist(u), ncol = 3)))
eta_sim0 <- apply(result[[5]], 1, function (u) matrix(unlist(u), ncol = 1))
cov_sim0 <- t(apply(result[[6]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
stop_sim0 <- t(apply(result[[7]], 1, function (u) matrix(unlist(u), ncol = (sample_size-2))))


set.seed(1234)

n_dose <- sample_size + 1
xdose_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
ydose_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
cov_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
stop_sim <- matrix(NA, ncol = (sample_size-1), nrow = n_sim)
dlt_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
eta_sim <- matrix(NA, ncol = 1, nrow = n_sim)
time_sim <- matrix(NA, ncol = 1, nrow = n_sim)
rho_sim <- matrix(NA, ncol = 3, nrow = n_sim)
alpha_sim <- matrix(NA, ncol = nc, nrow = n_sim)
alpha_sim[, 1] <- alpha

registerDoParallel(ncores)
result <-
  foreach(i = 1:n_sim,
          .combine='comb',
          .multicombine=TRUE,
          .errorhandling= "remove",
          .init=list(list(), list(), list(), list(), list(), list(), list()),
          .packages='rjags') %dopar% {
            
            ## Binary Covariate
            cov <- matrix(rep(1, sample_size), ncol = 2)
            
            x.index <- 1
            y.index <- 2
            
            dlt <- matrix(NA, nrow = nc, ncol = npatients)
            dlt[1, ] <- 0
            xdose <- matrix(NA, nrow = nc, ncol = npatients)
            xdose[1, ] <- 0
            ydose <- matrix(NA, nrow = nc, ncol = npatients)
            ydose[1, ] <- 0
            stop_rule <- matrix(NA, nrow = (nc - 1), ncol = 2)
            
            for (j in 2:nc) {
              response <- data.frame(npatients = 1,
                                     dlt = as.numeric(t(dlt[1:(j-1), ])))
              design_matrix <- data.frame(intercept = 1,
                                          x = as.numeric(t(xdose[1:(j-1), ])),
                                          y = as.numeric(t(ydose[1:(j-1), ])), 
                                          xy = as.numeric(t(xdose[1:(j-1), ]))*
                                            as.numeric(t(ydose[1:(j-1), ])))
              
              x.index <- setdiff(1:2, x.index)
              y.index <- setdiff(1:2, x.index)
              
              my_data <- list(response = response, design_matrix = design_matrix,
                              theta = theta, alpha = alpha_sim[i, (j-1)],
                              limits_xdose = limits_xdose, 
                              limits_ydose = limits_ydose, 
                              rho_prior = rho_prior, 
                              
                              interaction_prior = interaction_prior,
                              n_adapt = n_adapt, burn_in = burn_in, 
                              n_mcmc = n_mcmc, n_thin = n_thin, 
                              n_chains = n_chains,
                              y.index = y.index, x.index = x.index)
              
              dose <- nd_ewoc_extended2d(my_data, 
                                         x = xdose[(j-1), ],
                                         y = ydose[(j-1), ],
                                         cov = cov[(j-1), ],
                                         x.index = x.index,
                                         y.index = y.index,
                                         deltax = deltax,
                                         deltay = deltay)
              
              xdose[j, x.index] <- dose$next_xdose
              xdose[j, y.index] <- xdose[(j-1), y.index]
              
              ydose[j, y.index] <- dose$next_ydose
              ydose[j, x.index] <- ydose[(j-1), x.index]
              
              alpha_sim[i, j] <- ifelse(alpha_sim[i, (j-1)] < 0.49, 
                                        alpha_sim[i, (j-1)] + 0.05, 0.5)
              
              dlt[j, ] <- response_binary2d(x = xdose[j, ], y = ydose[j, ], 
                                            cov = cov[j, ],
                                            rho = true_rho, eta = true_eta,
                                            model = model, theta = theta)
              
              stop_rule[(j-1), ] <- mean(abs(dose$pdlt - theta) >stop01) 
            }
            
            list(xdose = as.numeric(t(xdose)), 
                 ydose = as.numeric(t(ydose)), 
                 dlt = as.numeric(t(dlt)), 
                 rho_sim = apply(dose$rho, 2, median), 
                 eta_sim = median(dose$eta),
                 cov = as.numeric(t(cov)),
                 stop_rule =  as.numeric(t(stop_rule)))
          }
stopImplicitCluster()

xdose_sim1 <- t(apply(result[[1]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
ydose_sim1 <- t(apply(result[[2]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
dlt_sim1 <- t(apply(result[[3]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
rho_sim1 <- t(apply(result[[4]], 1, function (u) matrix(unlist(u), ncol = 3)))
eta_sim1 <- apply(result[[5]], 1, function (u) matrix(unlist(u), ncol = 1))
cov_sim1 <- t(apply(result[[6]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
stop_sim1 <- t(apply(result[[7]], 1, function (u) matrix(unlist(u), ncol = (sample_size-2))))

v1 <- nrow(xdose_sim1) != n_sim
v0 <- nrow(xdose_sim0) != n_sim

if (v1){
  aux <- matrix(NA, nrow = (n_sim - nrow(xdose_sim1) ), ncol = sample_size)
  xdose_sim1 <- rbind(xdose_sim1, aux)
  ydose_sim1 <- rbind(ydose_sim1, aux)
  dlt_sim1 <- rbind(dlt_sim1, aux)
  rho_sim1 <- rbind(rho_sim1, aux[, 1:3])
  cov_sim1 <- rbind(cov_sim1, aux)
  eta_sim1 <- rbind(eta_sim1, aux[, 1, drop = FALSE])
  stop_sim1 <- rbind(stop_sim1, aux[, 1:(sample_size - 2)])
}

if (v0){
  aux <- matrix(NA, nrow = (n_sim - nrow(xdose_sim0)), ncol = sample_size)
  xdose_sim0 <- rbind(xdose_sim0, aux)
  ydose_sim0 <- rbind(ydose_sim0, aux)
  dlt_sim0 <- rbind(dlt_sim0, aux)
  rho_sim0 <- rbind(rho_sim0, aux[, 1:3])
  cov_sim0 <- rbind(cov_sim0, aux)
  eta_sim0 <- rbind(eta_sim0, aux[, 1, drop = FALSE])
  stop_sim0 <- rbind(stop_sim0, aux[, 1:(sample_size - 2)])
}

xdose_sim <- cbind(xdose_sim0, xdose_sim1)
ydose_sim <- cbind(ydose_sim0, ydose_sim1)
dlt_sim <- cbind(dlt_sim0, dlt_sim1)
rho_sim <- cbind(rho_sim0, rho_sim1)
eta_sim <- cbind(eta_sim0, eta_sim1)
cov_sim <- cbind(cov_sim0, cov_sim1)
stop_sim <- cbind(stop_sim0, stop_sim1)

save.image(file = paste0("s0", scenario, "_", model, "_st.Rdata"))
             
             
  