library(parallel)
library(doParallel)
library(R2OpenBUGS)
library(rjags)

############## Parameters
#Number of cores
ncores <- 15

# 01, 02, 03
scenario <- 01

# Trial
sample_size <- 40
npatients <- 2
nc <- sample_size/npatients

n_sim <- 1200
n_adapt <- 5000
burn_in <- 10000
n_chains <- 1
n_thin <- 1
n_mcmc <- 3000

rho_prior <- matrix(1, ncol = 2, nrow = 4)
interaction_prior <- matrix(c(0.8, 0.0384), ncol = 2, nrow = 1)

theta <- 0.33
alpha <- 0.25
limits_xdose <- c(0, 1)
limits_ydose <- c(0, 1)
deltax <- 0.2
deltay <- 0.2

stop01 <- 0.1
stop02 <- 0.3
min.patients <- 1

# Loading scenario
source(paste0("scenario0", scenario, ".R"))

############## Auxiliar functions

round_nearest <- function(dose, grid){
  
  dif <- abs(dose - grid)
  index <- ifelse(length(which.min(dif)) == 0, 1, which.min(dif))
  out <- grid[index]
  return(out)
}

scenario_grid <- function(ind) {
  
  if (ind == 1){
    nx <- 5
    ny <- 5
    
    x = rep(seq(0, 1, length.out = nx), each = ny)
    y = rep(seq(0, 1, length.out = ny), nx)
    cov = rep(c(0, 1), each = ny*nx)
    grid <- data.frame(x = x, y = y, cov = cov, p = NA)
    
    grid$p <- c(c(0.001, 0.05, 0.13, 0.2, 0.25,
                  0.05, 0.13, 0.16, 0.26, 0.33,
                  0.13, 0.18, 0.24, 0.33, 0.4, 
                  0.2,  0.28, 0.33, 0.43, 0.48,
                  0.27, 0.33, 0.39, 0.55, 0.7),
                c(0.001, 0.05, 0.13, 0.2, 0.25,
                  0.05, 0.13, 0.16, 0.26, 0.33,
                  0.13, 0.18, 0.24, 0.33, 0.4, 
                  0.2,  0.28, 0.33, 0.43, 0.48,
                  0.27, 0.33, 0.39, 0.55, 0.7) + 0.2)
    
    #Control jump
    delta1x <- 1/(nx - 1)
    delta1y <- 1/(ny - 1)
  }
  
  if (ind == 2){
    nx <- 5
    ny <- 5
    
    x = rep(seq(0, 1, length.out = nx), each = ny)
    y = rep(seq(0, 1, length.out = ny), nx)
    cov = rep(c(0, 1), each = ny*nx)
    grid <- data.frame(x = x, y = y, cov = cov, p = NA)
    
    grid$p <- c(c(0.01, 0.05, 0.14, 0.15, 0.17,
                  0.03, 0.08, 0.15, 0.16, 0.18,
                  0.14, 0.15, 0.17, 0.18, 0.20, 
                  0.15, 0.16, 0.19, 0.33, 0.35,
                  0.17, 0.18, 0.21, 0.38, 0.40),
                c(0.01, 0.05, 0.14, 0.15, 0.17,
                  0.03, 0.08, 0.15, 0.16, 0.18,
                  0.14, 0.15, 0.17, 0.18, 0.20, 
                  0.15, 0.16, 0.19, 0.33, 0.35,
                  0.17, 0.18, 0.21, 0.38, 0.40) + 0.3)
    
    #Control jump
    delta1x <- 1/(nx - 1)
    delta1y <- 1/(ny - 1)
  }
  
  if (ind == 3){
    nx <- 5
    ny <- 5
    
    x = rep(seq(0, 1, length.out = nx), each = ny)
    y = rep(seq(0, 1, length.out = ny), nx)
    cov = rep(c(0, 1), each = ny*nx)
    grid <- data.frame(x = x, y = y, cov = cov, p = NA)
    
    grid$p <- c(c(0.03, 0.03, 0.05, 0.08, 0.1,
                  0.03, 0.03, 0.10, 0.13, 0.15,
                  0.06, 0.10, 0.15, 0.18, 0.2, 
                  0.08, 0.13, 0.18, 0.33, 0.35,
                  0.10, 0.15, 0.20, 0.38, 0.40),
                c(0.03, 0.03, 0.14, 0.2, 0.25,
                  0.03, 0.03, 0.18, 0.25, 0.3,
                  0.14, 0.15, 0.20, 0.30, 0.35, 
                  0.20, 0.25, 0.30, 0.35, 0.4,
                  0.25, 0.30, 0.35, 0.40, 0.45) + 0.3)
    
    #Control jump
    delta1x <- 1/(nx - 1)
    delta1y <- 1/(ny - 1)
  }
  
  list(grid = grid, deltax = deltax, deltay = deltay, nx = nx, ny = ny) 
}

comb <- function(...) {
  mapply('rbind', ..., SIMPLIFY=FALSE)
}

response_discrete_binary2d <- function(x, y, cov, grid){
  
  aux_p <- function(aux, grid){
    p <- grid[which(grid$x == aux[1] & grid$y == aux[2] & grid$cov == aux[3]), 4]
    return(p)
  }
  
  aux <- cbind(x, y, cov)
  p <- apply(aux, 1, aux_p, grid = grid)
  out <- rbinom(length(x), 1, p)
  
  return(out)
}


pdlt_binary2d <- function(cov, x, y, rho, eta, model){
  
  aux_pdlt <- function(parm, x, y, cov, model){
    rho <- parm[1:4]
    eta <- parm[5]
    
    beta <- rep(NA, 5)
    beta[1] <- qlogis(rho[1])
    beta[2] <- qlogis(rho[2]) - qlogis(rho[1])
    beta[3] <- qlogis(rho[3]) - qlogis(rho[1])
    beta[4] <- qlogis(rho[4]) - qlogis(rho[2])
    beta[5] <- eta
    design <- cbind(1, x, y, cov, x*y)
    
    p <- plogis(design%*%beta)
    
    return(p)
  }
  
  parm <- cbind(rho, eta)
  
  out <- apply(parm, 1, aux_pdlt, 
               x = x, y = y, cov = cov, model = model)
  
  return(out)
}

### JAGS
binary2d_jags <- function(data) {
  
  theta <- data$theta
  lbx <- - data$limits_xdose[1]/(data$limits_xdose[2] - data$limits_xdose[1])
  lby <- - data$limits_ydose[1]/(data$limits_ydose[2] - data$limits_ydose[1])
  rho_prior <- data$rho_prior
  interaction_prior <- data$interaction_prior
  design_matrix <- data$design_matrix
  dlt <- data$response[, 2]
  npatients <- data$response[, 1]
  
  x.index <- data$x.index
  y.index<- data$y.index
  
  n_adapt <- data$n_adapt
  burn_in <- data$burn_in
  n_mcmc <- data$n_mcmc
  n_thin <- data$n_thin
  n_chains <- data$n_chains
  
  # JAGS model function
  jfun <- "model {
  
  for(i in 1:nobs) {
  dlt[i] ~ dbin(p[i], 1)
  p[i] <- ifelse(1/(1 + exp(-lp[i])) == 1, 0.99, 1/(1 + exp(-lp[i])))
  lp[i] <- inprod(design_matrix[i, ], beta)
  }
  
  beta[1] <- logit(rho[1])
  beta[2] <- logit(rho[2]) - logit(rho[1])
  beta[3] <- logit(rho[3]) - logit(rho[1])
  beta[4] <- logit(rho[4]) - logit(rho[2])
  beta[5] <- eta
  
  rho[1] <- limit*v[1]
  v[1] ~ dbeta(rho_prior[1, 1], rho_prior[1, 2])
  limit <- min(rho[2], rho[3])
  
  rho[2] <- v[2]
  rho[3] <- v[3]
  rho[4] <- v[4]
  
  v[2] ~ dbeta(rho_prior[2, 1], rho_prior[2, 2])
  v[3] ~ dbeta(rho_prior[3, 1], rho_prior[3, 2])
  v[4] ~ dbeta(rho_prior[4, 1], rho_prior[4, 2])
  eta ~ dgamma(interaction_prior[1, 1], interaction_prior[1, 2])
}"
  
  data_base <- list('dlt' = dlt, 'design_matrix' = design_matrix, 
                    'nobs' = length(dlt), 'rho_prior' = rho_prior,
                    'interaction_prior' = interaction_prior)
  
  inits <- function() {
    v <- rep(NA, 3)
    v[1] <- rbeta(1, rho_prior[1, 1], rho_prior[1, 2])
    v[2] <- rbeta(1, rho_prior[2, 1], rho_prior[2, 2])
    v[3] <- rbeta(1, rho_prior[3, 1], rho_prior[3, 2])
    v[4] <- rbeta(1, rho_prior[4, 1], rho_prior[4, 2])
    eta <- rgamma(1, interaction_prior[1, 1], interaction_prior[1, 2])
    out <- list(v = v, eta = eta)
    return(out)
  }

  # Calling JAGS
  j <- jags.model(textConnection(jfun),
                  data = data_base,
                  inits = inits(),
                  n.chains = n_chains,
                  n.adapt = n_adapt)
  update(j, burn_in)
  result <- coda.samples(j, variable.names = c("rho", "eta"),
                         n.iter = n_mcmc, thin = n_thin,
                         n.chains = n_chains)[[1]]
  rho <- result[, 2:5]
  eta <- result[, 1]
  
  out <- list(rho = rho, eta = eta)
  
  return(out)
}


standard_dose <- function(dose, min_dose, max_dose) {
  
  out <- (dose - min_dose)/(max_dose - min_dose)
  return(out)
}

nd_ewoc_binary2d <- function(data, x, y, cov, x.index, y.index, deltax, deltay){
  
  mcmc_sample <- binary2d_jags(data) 
  rho <- mcmc_sample$rho
  eta <- mcmc_sample$eta
  lbx <- - data$limits_xdose[1]/(data$limits_xdose[2] - data$limits_xdose[1])
  lby <- - data$limits_ydose[1]/(data$limits_ydose[2] - data$limits_ydose[1])
  
  ymtd <- ((qlogis(theta) - qlogis(rho[, 1])) - 
    (qlogis(rho[, 2]) - qlogis(rho[, 1]))*x[x.index] - 
      (qlogis(rho[, 4]) - qlogis(rho[, 2]))*cov[x.index])/
    ((qlogis(rho[, 3]) - qlogis(rho[, 1])) + eta*x[x.index])
  ymtd <- ymtd[ymtd > lby]
  
  next_ydose <- quantile(ymtd, probs = data$alpha)
  if ((next_ydose - y[x.index]) > deltay)
    next_ydose <- y[x.index] + deltay
  
  if (next_ydose > 1)
    next_ydose <- 1
  if (next_ydose < 0)
    next_ydose <- 0
  
  xmtd <- ((qlogis(theta) - qlogis(rho[, 1])) - 
    (qlogis(rho[, 3]) - qlogis(rho[, 1]))*y[y.index] - 
      (qlogis(rho[, 4]) - qlogis(rho[, 2]))*cov[y.index])/
    ((qlogis(rho[, 2]) - qlogis(rho[, 1])) + eta*y[y.index])
  xmtd <- xmtd[xmtd > lbx]
  
  next_xdose <- quantile(xmtd, probs = data$alpha)
  if ((next_xdose - x[y.index]) > deltax)
    next_xdose <- x[y.index] + deltax
  
  if (next_xdose > 1)
    next_xdose <- 1
  if (next_xdose < 0)
    next_xdose <- 0
  
  post_low <- apply(matrix(cov, nrow = length(cov)), 1, pdlt_binary2d, 
                x = 0, y = 0, rho = rho, eta = eta, 
                        model = "logistic")
  
  out <- list(next_ydose = next_ydose, next_xdose = next_xdose,
              ymtd = ymtd, xmtd = xmtd, rho = rho, eta = eta, 
              post_low = post_low)
}

### Simulation
aux <- scenario_grid(scenario)
deltax <- aux$deltax
deltay <- aux$deltay
nx <- aux$nx
ny <- aux$ny
grid <- aux$grid

set.seed(1234)
n_dose <- sample_size + 1
xdose_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
ydose_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
cov_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
stop_sim <- matrix(NA, ncol = (sample_size-1), nrow = n_sim)
dlt_sim <- matrix(NA, ncol = sample_size, nrow = n_sim)
eta_sim <- matrix(NA, ncol = 1, nrow = n_sim)
time_sim <- matrix(NA, ncol = 1, nrow = n_sim)
rho_sim <- matrix(NA, ncol = 3, nrow = n_sim)
alpha_sim <- matrix(NA, ncol = nc, nrow = n_sim)
alpha_sim[, 1] <- alpha

registerDoParallel(ncores)
result <-
  foreach(i = 1:n_sim,
          .combine='comb',
          .multicombine=TRUE,
          .errorhandling= "remove",
          .init=list(list(), list(), list(), list(), list(), list(), list(), list()),
          .packages='rjags') %dopar% {
            
            ## Binary Covariate
            temp <- c(rep(0, sample_size/2), rep(1, sample_size/2))
            cov <- matrix(sample(temp, size = sample_size, replace = FALSE), ncol = 2)
            
            x.index <- 2
            y.index <- 1
            
            dlt <- matrix(NA, nrow = nc, ncol = npatients)
            dlt[1, ] <- 0
            xdose <- matrix(NA, nrow = nc, ncol = npatients)
            xdose[1, ] <- 0
            ydose <- matrix(NA, nrow = nc, ncol = npatients)
            ydose[1, ] <- 0
            stop_rule <- matrix(NA, nrow = (nc - 1), ncol = 2)
            
            for (j in 2:nc) {
              response <- data.frame(npatients = 1,
                                     dlt = as.numeric(t(dlt[1:(j-1), ])))
              design_matrix <- data.frame(intercept = 1,
                                          x = as.numeric(t(xdose[1:(j-1), ])),
                                          y = as.numeric(t(ydose[1:(j-1), ])), 
                                          cov = as.numeric(t(cov[1:(j-1), ])),
                                          xy = as.numeric(t(xdose[1:(j-1), ]))*
                                            as.numeric(t(ydose[1:(j-1), ])))
              
              x.index <- setdiff(1:2, x.index)
              y.index <- setdiff(1:2, x.index)
              
              my_data <- list(response = response, design_matrix = design_matrix,
                              theta = theta, alpha = alpha_sim[i, (j-1)],
                              limits_xdose = limits_xdose, 
                              limits_ydose = limits_ydose, 
                              rho_prior = rho_prior, 
                              interaction_prior = interaction_prior,
                              n_adapt = n_adapt, burn_in = burn_in, 
                              n_mcmc = n_mcmc, n_thin = n_thin, 
                              n_chains = n_chains,
                              y.index = y.index, x.index = x.index)
              
              dose <- nd_ewoc_binary2d(my_data, 
                                       x = xdose[(j-1), ],
                                       y = ydose[(j-1), ],
                                       cov = cov[j, ],
                                       x.index = x.index,
                                       y.index = y.index,
                                       deltax = deltax,
                                       deltay = deltay)
              
              count <- table(cov[1:(j-1), ])
              cond <- names(count) == cov[j, x.index]
              if (any(cond)){
                if (as.numeric(count[cond]) >= min.patients){
                  ydose[j, x.index] <- round_nearest(dose$next_ydose, 
                                                     grid[cov == cov[j, x.index], 2])
                } else {
                  ydose[j, x.index] <- ydose[(j-1), x.index]
                }
              } else {
                ydose[j, x.index] <- 0
              }
              ydose[j, y.index] <- ydose[(j-1), y.index]
              
              cond <- names(count) == cov[j, y.index]
              
              if (any(cond)){
                if (as.numeric(count[cond]) >= min.patients){
                  xdose[j, y.index] <- round_nearest(dose$next_xdose, 
                                                     grid[cov == cov[j, y.index], 1])
                } else {
                  xdose[j, y.index] <- xdose[(j-1), y.index]
                }
              } else {
                xdose[j, y.index] <- 0
              }
              xdose[j, x.index] <- xdose[(j-1), x.index]
              
              alpha_sim[i, j] <- ifelse(alpha_sim[i, (j-1)] < 0.49, 
                                        alpha_sim[i, (j-1)] + 0.05, 0.5)
              
              dlt[j, ] <- response_discrete_binary2d(x = xdose[j, ], y = ydose[j, ], 
                                                     cov = cov[j, ], grid = grid)
              
              stop_rule[(j-1), ] <- colMeans(dose$post_low > theta + stop01) 
            }
            
            j <- nc + 1
            response <- data.frame(npatients = 1,
                                   dlt = as.numeric(t(dlt[1:(j-1), ])))
            design_matrix <- data.frame(intercept = 1,
                                        x = as.numeric(t(xdose[1:(j-1), ])),
                                        y = as.numeric(t(ydose[1:(j-1), ])), 
                                        cov = as.numeric(t(cov[1:(j-1), ])),
                                        xy = as.numeric(t(xdose[1:(j-1), ]))*
                                          as.numeric(t(ydose[1:(j-1), ])))
            
            x.index <- setdiff(1:2, x.index)
            y.index <- setdiff(1:2, x.index)
            
            my_data <- list(response = response, design_matrix = design_matrix,
                            theta = theta, alpha = alpha_sim[i, (j-1)],
                            limits_xdose = limits_xdose, 
                            limits_ydose = limits_ydose, 
                            rho_prior = rho_prior, 
                            interaction_prior = interaction_prior,
                            n_adapt = n_adapt, burn_in = burn_in, 
                            n_mcmc = n_mcmc, n_thin = n_thin, 
                            n_chains = n_chains,
                            y.index = y.index, x.index = x.index)
            
            dose <- nd_ewoc_binary2d(my_data, 
                                     x = xdose[(j-1), ],
                                     y = ydose[(j-1), ],
                                     cov = cov[(j-1), ],
                                     x.index = x.index,
                                     y.index = y.index,
                                     deltax = deltax,
                                     deltay = deltay)
            
            post_dlt <- rep(NA, nrow(grid))
            for (k in 1:nrow(grid)){
              post_dlt[k] <- 
                mean(pdlt_binary2d(x = grid[k, 1], y = grid[k, 2], 
                                       cov = grid[k, 3],
                                         rho = dose$rho, 
                                         eta = dose$eta, 
                                         model = "logistic") > theta + 0.1|
                       pdlt_binary2d(x = grid[k, 1], y = grid[k, 2], 
                                     cov = grid[k, 3],
                                     rho = dose$rho, 
                                     eta = dose$eta, 
                                     model = "logistic") < theta - 0.1)
            }

  list(xdose = as.numeric(t(xdose)), 
       ydose = as.numeric(t(ydose)), 
       dlt = as.numeric(t(dlt)), 
       rho_sim = apply(dose$rho, 2, median), 
       eta_sim = median(dose$eta),
       cov = as.numeric(t(cov)),
       stop_rule =  as.numeric(t(stop_rule)),
       post_dlt = post_dlt)
          }
stopImplicitCluster()

xdose_sim <- t(apply(result[[1]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
ydose_sim <- t(apply(result[[2]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
dlt_sim <- t(apply(result[[3]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
rho_sim <- t(apply(result[[4]], 1, function (u) matrix(unlist(u), ncol = 4)))
eta_sim <- apply(result[[5]], 1, function (u) matrix(unlist(u), ncol = 1))
cov_sim <- t(apply(result[[6]], 1, function (u) matrix(unlist(u), ncol = sample_size)))
stop_sim <- t(apply(result[[7]], 1, function (u) matrix(unlist(u), ncol = (sample_size-2))))
aux <- result[[8]] 
aux <- aux[1:1000, ]
post_dlt_sim <- data.frame(trial = rep(1:nrow(xdose_sim), each = nrow(grid)),
                           x = rep(grid$x, nrow(xdose_sim)), 
                           y = rep(grid$y, nrow(ydose_sim)),
                           cov = rep(grid$cov, nrow(xdose_sim)),
                           p =  as.numeric(t(aux)))

if (nrow(xdose_sim) > 1000){
  
  xdose_sim <- xdose_sim[1:1000, ]
  ydose_sim <- ydose_sim[1:1000, ]
  dlt_sim <- dlt_sim[1:1000, ]
  rho_sim <- rho_sim[1:1000, ]
  eta_sim <- eta_sim[1:1000]
  cov_sim <- cov_sim[1:1000, ]
  stop_sim <- stop_sim[1:1000, ]
  post_dlt_sim <- post_dlt_sim[1:1000, ]
  }

save.image(file = paste0("s0", scenario, "_discrete_wc.Rdata"))
             
             
  